/*
 * tpdfview - src/gtk-tpdfv.c
 * Copyright (C) 2011 Sebastian 'tokkee' Harl <sh@tokkee.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * This module provides a Gtk widget to display PDF files.
 */

#include "gtk-tpdfv.h"

#include <assert.h>

#include <errno.h>

#include <cairo.h>
#include <gtk/gtk.h>

#include <poppler.h>

#include <stdlib.h>
#include <string.h>

#include <sys/stat.h>

#define TPDFV_MIN(a, b) ((a) <= (b) ? (a) : (b))
#define TPDFV_MAX(a, b) ((a) >= (b) ? (a) : (b))

typedef enum {
	TPDFV_ZOOM_CUSTOM = 0,
	TPDFV_ZOOM_WIDTH,
	TPDFV_ZOOM_HEIGHT,
	TPDFV_ZOOM_FIT,
} tpdfv_zoommode_t;

typedef struct {
	char *filename;

	time_t mtime;

	PopplerDocument *doc;
	PopplerPage     *current_page;

	int current_page_no;
	int total_pages;

	tpdfv_zoommode_t zoom_mode;
	double zoom_factor;

	double delta_x;
	double delta_y;
} gtk_tpdfv_t;

GType gtk_tpdfv_get_type(void);
G_DEFINE_TYPE(GtkTPDFV, gtk_tpdfv, GTK_TYPE_DRAWING_AREA);

#define GTK_TPDFV_GET_PRIVATE(obj) \
	(G_TYPE_INSTANCE_GET_PRIVATE ((obj), GTK_TYPE_TPDFV, gtk_tpdfv_t))

/*
 * Private helper functions.
 */

static void
tpdfv_open(gtk_tpdfv_t *pdf)
{
	GError *err = NULL;

	struct stat statbuf;

	if ((! pdf) || (! pdf->filename))
		return;

	memset(&statbuf, 0, sizeof(statbuf));
	if (stat(pdf->filename + strlen("file://"), &statbuf) != 0) {
		char errbuf[1024];
		strerror_r(errno, errbuf, sizeof(errbuf));
		fprintf(stderr, "Failed to access PDF: %s.\n", errbuf);
		return;
	}

	pdf->mtime = statbuf.st_mtime;

	pdf->doc = poppler_document_new_from_file(pdf->filename,
			/* password = */ NULL, &err);
	if (! pdf->doc) {
		fprintf(stderr, "Failed to open PDF: %s.\n", err->message);
		return;
	}

	pdf->total_pages = poppler_document_get_n_pages(pdf->doc);
	if (pdf->current_page_no >= pdf->total_pages)
		pdf->current_page_no = pdf->total_pages - 1;
} /* tpdfv_open */

static void
tpdfv_close(gtk_tpdfv_t *pdf)
{
	if (! pdf)
		return;

	g_object_unref(pdf->doc);
	pdf->doc = NULL;
	pdf->current_page = NULL;
} /* tpdfv_close */

static gboolean
tpdfv_init(gtk_tpdfv_t *pdf, const char *filename)
{
	if (strstr(filename, "://"))
		pdf->filename = strdup(filename);
	else {
		size_t len = strlen("file://") + strlen(filename);
		pdf->filename = (char *)malloc(len + 1);
		if (pdf->filename) {
			*pdf->filename = '\0';
			strncat(pdf->filename, "file://", len);
			strncat(pdf->filename, filename, len - strlen("file://"));
		}
	}

	/* XXX: error reporting mechanism */

	if (! pdf->filename) {
		char errbuf[1024];
		strerror_r(errno, errbuf, sizeof(errbuf));
		fprintf(stderr, "Failed to allocate string: %s.\n", errbuf);
		return FALSE;
	}

	pdf->doc = NULL;
	tpdfv_open(pdf);

	pdf->current_page_no = 0;
	pdf->current_page = poppler_document_get_page(pdf->doc,
			pdf->current_page_no);
	if (! pdf->current_page) {
		fprintf(stderr, "Failed to open first page of the document.\n");
		return FALSE;
	}

	pdf->zoom_mode   = TPDFV_ZOOM_CUSTOM;
	pdf->zoom_factor = 1.0;
	pdf->delta_x = pdf->delta_y = 0.0;
	return TRUE;
} /* tpdfv_init */

static void
tpdfv_clean(gtk_tpdfv_t *pdf)
{
	if (! pdf)
		return;

	tpdfv_close(pdf);
	pdf->current_page = NULL;
	free(pdf->filename);
	pdf->filename = NULL;
	return;
} /* tpdfv_clean */

static void
do_redraw(GtkWidget *widget)
{
	GdkRegion *region;
	gtk_tpdfv_t *pdf;

	pdf = GTK_TPDFV_GET_PRIVATE(widget);

	if ((! pdf->current_page) || (poppler_page_get_index(pdf->current_page)
			!= pdf->current_page_no)) {
		pdf->current_page = poppler_document_get_page(pdf->doc,
		pdf->current_page_no);

		if (! pdf->current_page) {
			fprintf(stderr, "Failed to open page %i of the document.\n",
					pdf->current_page_no + 1);
			return;
		}
	}

	region = gdk_drawable_get_clip_region(widget->window);
	gdk_window_invalidate_region(widget->window, region, TRUE);
	gdk_window_process_updates(widget->window, TRUE);

	gdk_region_destroy(region);
} /* do_redraw */

/*
 * Gtk+ class definition.
 */

static gboolean
gtk_tpdfv_expose(GtkWidget *tpdfv, GdkEventExpose *event)
{
	gtk_tpdfv_t *pdf;
	cairo_t *cr;

	struct stat statbuf;

	gdouble width, height;

	gdouble page_width  = 0.0;
	gdouble page_height = 0.0;

	double zoom_factor  = 1.0;
	double x, y;

	pdf = GTK_TPDFV_GET_PRIVATE(tpdfv);

	if (! pdf)
		return FALSE;

	memset(&statbuf, 0, sizeof(statbuf));
	if (stat(pdf->filename + strlen("file://"), &statbuf)) {
		char errbuf[1024];
		strerror_r(errno, errbuf, sizeof(errbuf));
		fprintf(stderr, "Failed to access PDF: %s.\n", errbuf);
		return FALSE;
	}

	if (statbuf.st_mtime > pdf->mtime)
		gtk_tpdfv_reload(tpdfv);

	cr = gdk_cairo_create(tpdfv->window);
	cairo_rectangle(cr, event->area.x, event->area.y,
			event->area.width, event->area.height);
	cairo_clip(cr);

	poppler_page_get_size(pdf->current_page, &page_width, &page_height);
	width  = (gdouble)tpdfv->allocation.width;
	height = (gdouble)tpdfv->allocation.height;

	/* zoom */
	if (pdf->zoom_mode == TPDFV_ZOOM_CUSTOM) {
		zoom_factor = pdf->zoom_factor;
	}
	else {
		if (pdf->zoom_mode == TPDFV_ZOOM_WIDTH) {
			zoom_factor = (double)(width / page_width);
		}
		else if (pdf->zoom_mode == TPDFV_ZOOM_HEIGHT) {
			zoom_factor = (double)(height / page_height);
		}
		else if (pdf->zoom_mode == TPDFV_ZOOM_FIT) {
			zoom_factor = (double)TPDFV_MIN(width / page_width,
					height / page_height);
		}
		else {
			assert(0);
		}
	}
	cairo_scale(cr, zoom_factor, zoom_factor);

	/* centered position / scrolling */
	x = (double)(width - page_width * zoom_factor) / 2.0 / zoom_factor;
	y = (double)(height - page_height * zoom_factor) / 2.0 / zoom_factor;

	cairo_translate(cr, x + pdf->delta_x, y + pdf->delta_y);

	/* draw empty white page */
	cairo_save(cr);

	cairo_set_source_rgba(cr, 1.0, 1.0, 1.0, 1.0);
	cairo_rectangle(cr, 0.0, 0.0, page_width, page_height);
	cairo_fill(cr);

	cairo_restore(cr);

	poppler_page_render(pdf->current_page, cr);

	cairo_destroy(cr);
	return FALSE;
} /* gtk_tpdfv_expose */

static void
gtk_tpdfv_destroy(GtkObject *obj)
{
	GtkTPDFV    *tpdfv = GTK_TPDFV(obj);
	gtk_tpdfv_t *pdf;

	pdf = GTK_TPDFV_GET_PRIVATE(tpdfv);
	tpdfv_clean(pdf);

	GTK_OBJECT_CLASS(gtk_tpdfv_parent_class)->destroy(obj);
} /* gtk_tpdfv_destroy */

static void
gtk_tpdfv_class_init(GtkTPDFVClass *class)
{
	GObjectClass   *goclass;
	GtkObjectClass *oclass;
	GtkWidgetClass *wclass;

	goclass = G_OBJECT_CLASS(class);
	g_type_class_add_private(goclass, sizeof(gtk_tpdfv_t));

	oclass = GTK_OBJECT_CLASS(class);
	oclass->destroy = gtk_tpdfv_destroy;

	wclass = GTK_WIDGET_CLASS(class);
	wclass->expose_event = gtk_tpdfv_expose;
} /* gtk_tpdfv_class_init */

static void
gtk_tpdfv_init(GtkTPDFV __attribute__((unused)) *tpdfv)
{
	/* nothing to do for now */
} /* gtk_tpdfv_init */

/*
 * Public API.
 */

GtkWidget *
gtk_tpdfv_new(const char *filename)
{
	GtkTPDFV    *tpdfv;
	gtk_tpdfv_t *pdf;

	tpdfv = g_object_new(GTK_TYPE_TPDFV, NULL);
	pdf   = GTK_TPDFV_GET_PRIVATE(tpdfv);

	if (TRUE != tpdfv_init(pdf, filename))
		return NULL;
	return GTK_WIDGET(tpdfv);
} /* gtk_tpdfv_new */

void
gtk_tpdfv_reload(GtkWidget *widget)
{
	gtk_tpdfv_t *pdf;

	pdf = GTK_TPDFV_GET_PRIVATE(widget);
	tpdfv_close(pdf);
	tpdfv_open(pdf);
	do_redraw(widget);
} /* gtk_tpdfv_reload */

void
gtk_tpdfv_page_up(GtkWidget *widget)
{
	gtk_tpdfv_t *pdf;

	pdf = GTK_TPDFV_GET_PRIVATE(widget);

	if (pdf->current_page_no)
		--pdf->current_page_no;
	do_redraw(widget);
} /* gtk_tpdfv_page_up */

void
gtk_tpdfv_page_down(GtkWidget *widget)
{
	gtk_tpdfv_t *pdf;

	pdf = GTK_TPDFV_GET_PRIVATE(widget);

	if (pdf->current_page_no < pdf->total_pages - 1)
		++pdf->current_page_no;
	do_redraw(widget);
} /* gtk_tpdfv_page_down */

void
gtk_tpdfv_first_page(GtkWidget *widget)
{
	gtk_tpdfv_t *pdf;

	pdf = GTK_TPDFV_GET_PRIVATE(widget);
	pdf->current_page_no = 0;
	do_redraw(widget);
} /* gtk_tpdfv_first_page */

void
gtk_tpdfv_last_page(GtkWidget *widget)
{
	gtk_tpdfv_t *pdf;

	pdf = GTK_TPDFV_GET_PRIVATE(widget);
	pdf->current_page_no = pdf->total_pages - 1;
	do_redraw(widget);
} /* gtk_tpdfv_last_page */

void
gtk_tpdfv_zoom_in(GtkWidget *widget)
{
	gtk_tpdfv_t *pdf;

	pdf = GTK_TPDFV_GET_PRIVATE(widget);
	pdf->zoom_factor *= 1.2;
	pdf->zoom_mode    = TPDFV_ZOOM_CUSTOM;
	do_redraw(widget);
} /* gtk_tpdfv_zoom_in */

void
gtk_tpdfv_zoom_out(GtkWidget *widget)
{
	gtk_tpdfv_t *pdf;

	pdf = GTK_TPDFV_GET_PRIVATE(widget);

	if (pdf->zoom_factor > DBL_EPSILON * 2.0) {
		pdf->zoom_factor /= 1.2;
		pdf->zoom_mode    = TPDFV_ZOOM_CUSTOM;
		do_redraw(widget);
	}
} /* gtk_tpdfv_zoom_out */

void
gtk_tpdfv_zoom_1(GtkWidget *widget)
{
	gtk_tpdfv_t *pdf;

	pdf = GTK_TPDFV_GET_PRIVATE(widget);
	pdf->zoom_factor = 1.0;
	pdf->zoom_mode   = TPDFV_ZOOM_CUSTOM;
	do_redraw(widget);
} /* gtk_tpdfv_zoom_1 */

void
gtk_tpdfv_zoom_width(GtkWidget *widget)
{
	gtk_tpdfv_t *pdf;

	pdf = GTK_TPDFV_GET_PRIVATE(widget);
	pdf->zoom_mode = TPDFV_ZOOM_WIDTH;
	do_redraw(widget);
} /* gtk_tpdfv_zoom_width */

void
gtk_tpdfv_zoom_height(GtkWidget *widget)
{
	gtk_tpdfv_t *pdf;

	pdf = GTK_TPDFV_GET_PRIVATE(widget);
	pdf->zoom_mode = TPDFV_ZOOM_HEIGHT;
	do_redraw(widget);
} /* gtk_tpdfv_zoom_width */

void
gtk_tpdfv_zoom_fit(GtkWidget *widget)
{
	gtk_tpdfv_t *pdf;

	pdf = GTK_TPDFV_GET_PRIVATE(widget);
	pdf->zoom_mode = TPDFV_ZOOM_FIT;
	do_redraw(widget);
} /* gtk_tpdfv_zoom_width */

void
gtk_tpdfv_scroll_up(GtkWidget *widget)
{
	gtk_tpdfv_t *pdf;

	pdf = GTK_TPDFV_GET_PRIVATE(widget);
	pdf->delta_y += 10.0;
	do_redraw(widget);
} /* gtk_tpdfv_scroll_up */

void
gtk_tpdfv_scroll_down(GtkWidget *widget)
{
	gtk_tpdfv_t *pdf;

	pdf = GTK_TPDFV_GET_PRIVATE(widget);
	pdf->delta_y -= 10.0;
	do_redraw(widget);
} /* gtk_tpdfv_scroll_down */

void
gtk_tpdfv_scroll_left(GtkWidget *widget)
{
	gtk_tpdfv_t *pdf;

	pdf = GTK_TPDFV_GET_PRIVATE(widget);
	pdf->delta_x += 10.0;
	do_redraw(widget);
} /* gtk_tpdfv_scroll_left */

void
gtk_tpdfv_scroll_right(GtkWidget *widget)
{
	gtk_tpdfv_t *pdf;

	pdf = GTK_TPDFV_GET_PRIVATE(widget);
	pdf->delta_x -= 10.0;
	do_redraw(widget);
} /* gtk_tpdfv_scroll_right */

/* vim: set tw=78 sw=4 ts=4 noexpandtab : */

